use std::collections::HashMap;

use crate::VARIABLE_START;
use crate::COMMENT;

pub fn preprocess(path : String) -> Vec<(usize, String)> {
    let contents = match std::fs::read_to_string(&path) {
        Ok(s) => s,
        Err(e) => {
            println!("ERROR: could not read file {}", path);
            println!("{}", e);
            std::process::exit(-1);
        }
    };

    let lines = get_lines(contents);

    let symbol_table = generate_symbol_table(&lines);

    let substituted_lines = substitute_symbols(symbol_table, lines);

    substituted_lines
}

fn get_lines(contents : String) -> Vec<(usize, String)> {
    let mut lines: Vec<(usize, String)> = Vec::new();

    for (i,line) in contents.lines().enumerate() {
        //let trimmed = String::from(line.trim());
        let mut trimmed = String::from(line);
        remove_whitespace(&mut trimmed);

        match remove_comments(trimmed) {
            None => continue,
            Some(trimmed) => lines.push((i, trimmed))
        }
    }

    lines
}

fn remove_whitespace(s: &mut String) {
    s.retain(|c| !c.is_whitespace());
}

fn remove_comments(line: String) -> Option<String> {
    let index = line.find(COMMENT);

    let stripped = match index {
        None => line,
        Some(i) => line[..i].to_string()
    };

    if stripped == "" {
        None
    }
    else {
        Some(stripped)
    }
}

fn generate_symbol_table(lines :  &Vec<(usize, String)>) -> HashMap<String, u32> {
    let mut symbol_table = prepopulate_symbol_table();

    //add all labels to table
    let mut pc = 0;
    for (i, line) in lines {
        if &line[0..1] == "(" {
            let close = match line.find(")") {
                None => {
                    println!("ERROR on line {}", i+1);
                    println!("Invalid Syntax");
                    std::process::exit(-1);
                },
                Some(i) => i,
            };

            let symbol = String::from(&line[1..close]);

            if !symbol_table.contains_key(&symbol) {
                symbol_table.insert(symbol, pc.try_into().unwrap());
            }
            else {
                println!("ERROR on line {}", i+1);
                println!("Duplicate label");
                std::process::exit(-1);
            }
        }
        else {
            pc += 1;
        }
    }

    //add all variables to table
    let mut next_address = VARIABLE_START;
    for (_, line) in lines {
        if &line[0..1] == "@" && !&line[1..2].chars().all(char::is_numeric)  {
            let symbol = String::from(&line[1..]);
            
            if !symbol_table.contains_key(&symbol) {
                symbol_table.insert(symbol, next_address);
                next_address += 1;
            }
        }
    }

    symbol_table
}

//prepopulate symbol table with predefined symbols
fn prepopulate_symbol_table() -> HashMap<String, u32> {
    let mut symbol_table = HashMap::new();

    symbol_table.insert(String::from("SP"), 0);
    symbol_table.insert(String::from("LCL"), 1);
    symbol_table.insert(String::from("ARG"), 2);
    symbol_table.insert(String::from("THIS"), 3);
    symbol_table.insert(String::from("THAT"), 4);

    symbol_table.insert(String::from("R0"), 0);
    symbol_table.insert(String::from("R1"), 1);
    symbol_table.insert(String::from("R2"), 2);
    symbol_table.insert(String::from("R3"), 3);
    symbol_table.insert(String::from("R4"), 4);
    symbol_table.insert(String::from("R5"), 5);
    symbol_table.insert(String::from("R6"), 6);
    symbol_table.insert(String::from("R7"), 7);
    symbol_table.insert(String::from("R8"), 8);
    symbol_table.insert(String::from("R9"), 9);
    symbol_table.insert(String::from("R10"), 10);
    symbol_table.insert(String::from("R11"), 11);
    symbol_table.insert(String::from("R12"), 12);
    symbol_table.insert(String::from("R13"), 13);
    symbol_table.insert(String::from("R14"), 14);
    symbol_table.insert(String::from("R15"), 15);

    symbol_table.insert(String::from("SCREEN"), 16384);
    symbol_table.insert(String::from("KBD"), 24576);

    symbol_table
}

fn substitute_symbols(symbol_table: HashMap<String, u32>, lines: Vec<(usize, String)>) -> Vec<(usize, String)> {
    let mut substituted_lines: Vec<(usize, String)> = Vec::new();

    match symbol_table.get("Sys.init") {
        Some(address) => println!("{}", address),
        None => println!("Not Found!"),
    };

    for (i, line) in lines {
        if &line[0..1] == "(" {
            continue;
        }
        else if &line[0..1] == "@" && !&line[1..2].chars().all(char::is_numeric)  {
            let symbol = String::from(&line[1..]);

            let substitution = match symbol_table.get(&symbol) {
                Some(address) => address,
                None => {
                    println!("ERROR on line {}", i+1);
                    println!("Unrecognized Symbol");
                    std::process::exit(-1);
                }
            };

            let mut new_line = String::from("@");
            new_line.push_str(&substitution.to_string());

            substituted_lines.push((i, new_line));
        }
        else {
            substituted_lines.push((i, line));
        }
    }

    substituted_lines
}