use crate::parser::Instruction;

#[derive(Debug)]
pub enum Code {
    ACode{val: String},
    CCode{dest: String, comp: String, jump: String},
}

impl Code {
    pub fn encode(&self) -> String {
        match self {
            Code::ACode{val} => format!("0{}", val),
            Code::CCode{comp, dest, jump} => format!("111{}{}{}", comp, dest, jump),
        } 
    }
}

pub fn translate(instructions: Vec<(usize,Instruction)>) -> Vec<Code> {
    let mut codes: Vec<Code> = Vec::new();

    for (i, instruction) in instructions {
        match translate_instruction(instruction) {
            None => {
                println!("ERROR on line {}", i+1);
                println!("Invalid Instruction");
                std::process::exit(-1);
            }
            Some(code) => codes.push(code),
        }
    }

    codes
}

fn translate_instruction(instruction: Instruction) -> Option<Code> {
    match instruction {
        Instruction::A{val} => {
            match encode_val(val) {
                None => None,
                Some(val) => Some(Code::ACode{val})
            }
        },
        Instruction::C{dest, comp, jump} => {
            let dest = match encode_dest(dest) {
                None => return None,
                Some(dest) => dest,
            };

            let comp = match encode_comp(comp) {
                None => return None,
                Some(comp) => comp,
            };

            let jump = match encode_jump(jump) {
                None => return None,
                Some(jump) => jump,
            };

            Some(Code::CCode{dest, comp, jump})
        }
    }
}

fn encode_val(val: i16) -> Option<String> {
    let mut bin = String::from("");

    let mut total = 0;
    let mut place = 15;

    while place > 0 {
        place -= 1;

        let x = i16::pow(2, place);
        if total + x <= val {
            bin.push_str("1");
            total += x;
        }
        else {
            bin.push_str("0");
        }
    }

    Some(bin)
}

fn encode_dest(dest: String) -> Option<String> {
    let mut code = String::from("");
    if dest.len() == 0 {
        code.push_str("000");
        Some(code)
    }
    else if dest.len() > 3 {
        None
    }
    else if dest.chars().all(|x| x == 'A' || x == 'D' || x == 'M') {
        // A Register
        if dest.contains("A") {
            code.push_str("1");
        }
        else {
            code.push_str("0");
        }

        //D register
        if dest.contains("D") {
            code.push_str("1");
        }
        else {
            code.push_str("0");
        }

        //M register
        if dest.contains("M") {
            code.push_str("1");
        }
        else {
            code.push_str("0");
        }

        Some(code)
    }
    else {
        None
    }
}

fn encode_comp(comp: String) -> Option<String> {
    match comp.as_str() {
        //a=0
        "0" => Some(String::from("0101010")),
        "1" => Some(String::from("0111111")),
        "-1" => Some(String::from("0111010")),
        "D" => Some(String::from("0001100")),
        "A" => Some(String::from("0110000")),
        "!D" => Some(String::from("0001101")),
        "!A" => Some(String::from("0110001")),
        "-D" => Some(String::from("0001111")),
        "-A" => Some(String::from("0110011")),
        "D+1" => Some(String::from("0011111")),
        "A+1" => Some(String::from("0110111")),
        "D-1" => Some(String::from("0001110")),
        "A-1" => Some(String::from("0110010")),
        "D+A" | "A+D" => Some(String::from("0000010")),
        "D-A" => Some(String::from("0010011")),
        "A-D" => Some(String::from("0000111")),
        "D&A" => Some(String::from("0000000")),
        "D|A" => Some(String::from("0010101")),
        //a=1
        "M" => Some(String::from("1110000")),
        "!M" => Some(String::from("1110001")),
        "-M" => Some(String::from("1110011")),
        "M+1" => Some(String::from("1110111")),
        "M-1" => Some(String::from("1110010")),
        "D+M" | "M+D" => Some(String::from("1000010")),
        "D-M" => Some(String::from("1010011")),
        "M-D" => Some(String::from("1000111")),
        "D&M" => Some(String::from("1000000")),
        "D|M" => Some(String::from("1010101")),
        _ => None,
    }
}

fn encode_jump(jump: String) -> Option<String> {
    match jump.as_str() {
        "" => Some(String::from("000")),
        "JGT" => Some(String::from("001")),
        "JEQ" => Some(String::from("010")),
        "JGE" => Some(String::from("011")),
        "JLT" => Some(String::from("100")),
        "JNE" => Some(String::from("101")),
        "JLE" => Some(String::from("110")),
        "JMP" => Some(String::from("111")),
        _ => None,
    }
}