use std::time::Instant;
use piston_window::*;
use hack16::{parser::Instruction};

pub const MEM_SIZE: usize = 24577;
pub const ROM_SIZE: usize= 32767;
pub const SCREEN_WIDTH: u32 = 512;
pub const SCREEN_HEIGHT: u32 = 256;
pub const SCREEN: u32 = 16384;
pub const KBD: usize = 24576;

pub struct Machine {
    clock: u64,
    tolerance: f64,
    timer: Instant,
    cycles: u64,
    a:i16,
    d:i16,
    pc:i16,
    out:i16,
    ram: [i16; MEM_SIZE],
    rom: Vec<Instruction>,
    window: PistonWindow,
    shift: bool,
    debug: bool,
}

impl Machine {
    pub fn new(clock: u64, tolerance: f64, fps: u64) -> Machine {
        let ram = [0; MEM_SIZE];
        let rom:Vec<Instruction> = Vec::new();

        let mut window: PistonWindow =
            WindowSettings::new("Simulated Screen", [SCREEN_WIDTH, SCREEN_HEIGHT])
            .exit_on_esc(false)
            .resizable(false)
            .automatic_close(true)
            .decorated(true)
            .vsync(false)
            .build().unwrap();

        window.set_ups(clock);
        window.set_ups_reset(0);
        window.set_max_fps(fps);

        Machine {
            clock,
            tolerance,
            timer: Instant::now(),
            cycles: 0,
            a:0,
            d: 0,
            pc: 0,
            out:0,
            ram,
            rom,
            window,
            shift: false,
            debug:false,
        }
    }

    pub fn load_rom(&mut self, instructions: Vec<(usize,Instruction)>) {
        if instructions.len() > ROM_SIZE {
            panic!("Too many instructions for rom");
        }

        for (_ ,instruction) in instructions {
            self.rom.push(instruction);
        }
    }

    pub fn set_ram(&mut self, address: usize, value: i16) {
        self.ram[address] = value;
    }

    pub fn start(&mut self) -> [i16; MEM_SIZE] {
        while let Some(event) = self.window.next() {
            match event {
                Event::Loop(e) => {
                    match e {
                        Loop::Render(..) => {
                            self.update_screen(&event);
                        },
                        Loop::Update(..) => {
                            //timings
                            if self.cycles == 0 {
                                self.timer = Instant::now();
                            }

                            self.cycles += 1;

                            let pc = usize::try_from(self.pc).expect("Invalid value in pc!");

                            if pc < self.rom.len() {
                                //get next instruction
                                let instruction = self.rom[pc].clone();
                
                                //execute instruction
                                self.execute(&instruction);
                                
                                //debug info
                                if self.debug {println!("instruction: {:?}", self.rom[pc]);}
                
                                if self.debug {
                                    self.print_state();
                                    println!("");
                                }
                            }
                            else {
                                panic!("pc out of range!");
                            }

                            //timings
                            if self.cycles == self.clock {
                                let elapsed = self.timer.elapsed().as_millis();

                                if elapsed > 1000 + ((1000.0 * self.tolerance) as u128) {
                                    println!("CLOCK RATE TOO HIGH! Last {} cycles completed in {} milliseconds (should be 1000)", self.clock, elapsed);
                                }
                                
                                self.cycles = 0;
                            }
                        },
                        _ => {},
                    }
                }
                Event::Input(input, ..) => {
                    if let Input::Button(button_args) = input {
                        if let Button::Keyboard(key) = button_args.button {
                            match button_args.state {
                                ButtonState::Press => {
                                    self.update_keycode(key);
                                },
                                ButtonState::Release => {
                                    match key {
                                        Key::LShift | Key::RShift => self.shift = false,
                                        _ => self.ram[KBD] = 0,
                                    }
                                }
                            }
                        }
                    }
                },
                _ => {},
            }
        }

        self.window.set_should_close(true);

        self.ram
    }

    pub fn execute(&mut self, instruction: &Instruction) {
        match instruction {
            Instruction::A{val} => {
                self.execute_a(*val);
            },
            Instruction::C{dest, comp, jump} => {
                self.execute_comp(comp);
                self.execute_dest(dest);
                self.execute_jump(jump);
            }
        }
    }

    pub fn print_state(&mut self) -> () {
        let addr =  usize::try_from(self.a).expect("Invalid A value!");
        if addr >= MEM_SIZE {
            println!("State: A: {}, D: {}, M: n/a, out: {}, pc: {}, SP: {}, KBD: {}", self.a, self.d, self.out, self.pc, self.ram[0], self.ram[KBD]);
        }
        else {
            println!("State: A: {}, D: {}, M: {}, out: {}, pc: {}, SP: {}, KBD: {}", self.a, self.d, self.ram[addr], self.out, self.pc, self.ram[0], self.ram[KBD]);
        }
    }

    pub fn set_debug(&mut self, val: bool) {
        self.debug = val;
    }    

    fn execute_a(&mut self, val: i16) {
        if val < 0 {
            panic!("A instruction value out of range! @[val] < 32768.");
        }
        else {
            self.a = val;
        }

        self.pc += 1;
    }

    fn execute_comp(&mut self, comp: &str) {
        let addr =  usize::try_from(self.a).expect("Invalid A value!");

        match comp {
            "0" => self.out = 0,
            "1" => self.out = 1,
            "-1" => self.out = -1,
            "D" => self.out = self.d,
            "A" => self.out = self.a,
            "!D" => self.out = !self.d,
            "!A" => self.out = !self.a,
            "-D" => self.out = 0 - self.d,
            "-A" => self.out = 0 - self.a,
            "D+1" => self.out = self.d + 1,
            "A+1" => self.out = self.a + 1,
            "D-1" => self.out = self.d - 1,
            "A-1" => self.out = self.a - 1,
            "D+A" | "A+D" => self.out = self.d + self.a,
            "D-A" => self.out = self.d - self.a,
            "A-D" => self.out = self.a - self.d,
            "D&A" => self.out = self.d & self.a,
            "D|A" => self.out = self.d | self.a,
            "M" => self.out = self.ram[addr],
            "!M" => self.out = !self.ram[addr],
            "-M" => self.out = 0 - self.ram[addr],
            "M+1" => self.out = self.ram[addr] + 1,
            "M-1" => self.out = self.ram[addr] - 1,
            "D+M" | "M+D" => self.out = self.d + self.ram[addr],
            "D-M" => self.out = self.d - self.ram[addr],
            "M-D" => self.out = self.ram[addr] - self.d,
            "D&M" => self.out = self.d & self.ram[addr],
            "D|M" => self.out = self.d & self.ram[addr],
            _ => panic!("Unrecognised comp instruction: \"{}\"!", comp),
        };
    }

    fn execute_dest(&mut self, dest: &str) {
        let addr =  usize::try_from(self.a).expect("Invalid A value!");
        let dest = String::from(dest);

        if dest.len() == 0 {
            return;
        }
        else if dest.len() > 3 {
            panic!("Unrecognised dest instruction: \"{}\"!", dest);
        }
        else if dest.chars().all(|x| x == 'A' || x == 'D' || x == 'M') {
            // A Register
            if dest.contains("A") {
                self.a = self.out;
            }
    
            //D register
            if dest.contains("D") {
                self.d = self.out;
            }
    
            //M register
            if dest.contains("M") {
                if self.a < MEM_SIZE as i16 {
                    self.ram[addr] = self.out;
                }
                else {
                    self.error("ERROR: attempting to write to M with invalid address (A).")
                }
            }
        }
        else {
            panic!("Unrecognised dest instruction: \"{}\"!", dest);
        }
    }

    fn execute_jump(&mut self, jump: &str) {
        self.pc += 1;

        match jump {
            "JGT" => {
                if self.out > 0 {self.pc = self.a;}
            },
            "JEQ" => {
                if self.out == 0 {self.pc = self.a;}
            },
            "JGE" => {
                if self.out >= 0 {self.pc = self.a;}
            },
            "JLT" => {
                if self.out < 0 {self.pc = self.a;}
            },
            "JNE" => {
                if self.out != 0 {self.pc = self.a;}
            },
            "JLE" => {
                if self.out <= 0 {self.pc = self.a;}
            },
            "JMP" => {
                self.pc = self.a;
            },
            "" => (),
            _ => panic!("Unrecognised jump instruction: \"{}\"!", jump),
        };
    }

    fn update_screen(&mut self, event: &Event) {
        self.window.draw_2d(event, |context, graphics, _device| {
            //clear display
            clear([1.0; 4], graphics);

            for row in 0..SCREEN_HEIGHT {
                for col in 0..SCREEN_WIDTH {
                    let word = self.ram[(SCREEN + (row * 32) + (col / 16)) as usize];
                    let bit = col % 16;

                    if word & (1 << bit) != 0 {
                        rectangle(color::BLACK,
                            [col as f64, row as f64, col as f64 + 1.0, row as f64 + 1.0],
                            context.transform,
                            graphics);
                    }
                }
            }
        });
    }

    fn update_keycode(&mut self, key: Key) {
        match key {
            Key::LShift | Key::RShift => self.shift = true,

            Key::Space => self.ram[KBD] = 32,

            Key::Comma => self.ram[KBD] = 44,
            Key::Period => self.ram[KBD] = 46,
            Key::Slash => self.ram[KBD] = 47,

            Key::D0 => self.ram[KBD] = 48,
            Key::D1 => self.ram[KBD] = 49,
            Key::D2 => self.ram[KBD] = 50,
            Key::D3 => self.ram[KBD] = 51,
            Key::D4 => self.ram[KBD] = 52,
            Key::D5 => self.ram[KBD] = 53,
            Key::D6 => self.ram[KBD] = 54,
            Key::D7 => self.ram[KBD] = 55,
            Key::D8 => self.ram[KBD] = 56,
            Key::D9 => self.ram[KBD] = 57,

            Key::Semicolon => self.ram[KBD] = 59,
            Key::Equals => self.ram[KBD] = 61,

            Key::A => self.ram[KBD] = 65,
            Key::B => self.ram[KBD] = 66,
            Key::C => self.ram[KBD] = 67,
            Key::D => self.ram[KBD] = 68,
            Key::E => self.ram[KBD] = 69,
            Key::F => self.ram[KBD] = 70,
            Key::G => self.ram[KBD] = 71,
            Key::H => self.ram[KBD] = 72,
            Key::I => self.ram[KBD] = 73,
            Key::J => self.ram[KBD] = 74,
            Key::K => self.ram[KBD] = 75,
            Key::L => self.ram[KBD] = 76,
            Key::M => self.ram[KBD] = 77,
            Key::N => self.ram[KBD] = 78,
            Key::O => self.ram[KBD] = 79,
            Key::P => self.ram[KBD] = 80,
            Key::Q => self.ram[KBD] = 81,
            Key::R => self.ram[KBD] = 82,
            Key::S => self.ram[KBD] = 83,
            Key::T => self.ram[KBD] = 84,
            Key::U => self.ram[KBD] = 85,
            Key::V => self.ram[KBD] = 86,
            Key::W => self.ram[KBD] = 87,
            Key::X => self.ram[KBD] = 88,
            Key::Y => self.ram[KBD] = 89,
            Key::Z => self.ram[KBD] = 90,

            Key::LeftBracket => self.ram[KBD] = 91,
            Key::Backslash => self.ram[KBD] = 92,
            Key::RightBracket => self.ram[KBD] = 93,

            //Special Characters (see table 4)
            Key::Return => self.ram[KBD] = 128,
            Key::Backspace => self.ram[KBD] = 129,
            Key::Left => self.ram[KBD] = 130,
            Key::Up => self.ram[KBD] = 131,
            Key::Right => self.ram[KBD] = 132,
            Key::Down => self.ram[KBD] = 133,
            Key::Home => self.ram[KBD] = 134,
            Key::End => self.ram[KBD] = 135,
            Key::PageUp => self.ram[KBD] = 136,
            Key::PageDown => self.ram[KBD] = 137,
            Key::Insert => self.ram[KBD] = 138,
            Key::Delete => self.ram[KBD] = 139,
            Key::Escape => self.ram[KBD] = 140,

            //F keys
            Key::F1 => self.ram[KBD] = 141,
            Key::F2 => self.ram[KBD] = 142,
            Key::F3 => self.ram[KBD] = 143,
            Key::F4 => self.ram[KBD] = 144,
            Key::F5 => self.ram[KBD] = 145,
            Key::F6 => self.ram[KBD] = 146,
            Key::F7 => self.ram[KBD] = 147,
            Key::F8 => self.ram[KBD] = 148,
            Key::F9 => self.ram[KBD] = 149,
            Key::F10 => self.ram[KBD] = 150,
            Key::F11 => self.ram[KBD] = 151,
            Key::F12 => self.ram[KBD] = 152,

            //NOT WORKING - probably due to UK keyboard layout
            Key::At => self.ram[KBD] = 63,
            Key::Backquote => self.ram[KBD] = 96,
            Key::Quotedbl => self.ram[KBD] = 34,
            Key::Hash => self.ram[KBD] = 35,
            Key::Quote => self.ram[KBD] = 39,

            _ => self.ram[KBD] = 0,
        }

        //if letter pressed and shift not held, convert to lower case (add 32)
        if self.ram[KBD] >= 65 && self.ram[KBD] <= 90 && !self.shift {
            self.ram[KBD] += 32;
        }
        //shift modifiers
        else if self.shift {
            self.ram[KBD] = match self.ram[KBD] {
                91 => 123,
                92 => 124,
                93 => 125,
                47 => 63,
                45 => 95,
                44 => 60,
                59 => 58,
                46 => 62,
                54 => 94,
                49 => 33,
                53 => 37,
                55 => 38,
                57 => 40,
                48 => 41,
                56 => 42,
                61 => 43,
                _ => self.ram[KBD],
            }
        }
    }

    fn error(&mut self, message: &str) -> ! {
        println!("{}", message);
        println!("Machine state at time of error:");
        self.print_state();
        std::process::exit(-1);
    }
}